// 002 u_c_search_memo
// 18 mar 2007

// -- (C) John COLIBRI 2007
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_search_memo;
  interface
    uses
        Classes // tShiftState
        , Graphics
        , Controls // tMouseButton
        , StdCtrls // tEdit, tLabel
        , ComCtrls // tRichEdit
        , ExtCtrls // tPanel
        ;

     type TSearchMemo= // Class(tPanel)
                       Class(tWinControl)
                         Private
                           m_c_panel: tPanel;
                             m_c_previous_button: tButton;
                             m_c_search_edit: tEdit;
                             m_c_next_button: tButton;
                             m_c_line_number_label: tLabel;
                             m_c_debug_label: tLabel;
                           m_c_richedit: tRichEdit;

                           m_search_string: String;
                           m_search_length: Integer;

                           m_text: String;
                           m_text_index: Integer;
                           m_text_length: Integer;

                           _m_did_hit_control: Boolean;

                           m_word_character_string: String;
                           m_word_character_set: Set Of Char;
                           m_whole_word_color, m_partial_word_color: Integer;

                           m_path, m_file_name: String;

                           procedure set_panel_color(p_color: Integer);
                           function f_memo_line_of_index(p_text_index: Integer): Integer;
                           procedure set_memo_top_line_index(p_top_line_index: Integer);
                           procedure search_next_occurence(p_color: Integer);
                           function f_is_whole_word: Boolean;

                           procedure handle_richedit_keypress(p_c_sender: TObject;
                               var pv_key: Char);
                           procedure handle_memo_and_search_edit_keydown(p_c_sender: TObject;
                               var pv_scan_code: Word; p_shift_state: TShiftState);

                           procedure handle_next_click(p_c_sender: tObject);
(*
                           // procedure handle_sort_click(p_c_sender: tObject);
                           procedure handle_memo_change(p_c_sender: tObject);
                           procedure handle_save_edit_keypress(p_c_sender: TObject;
                               var pv_key: Char);
*)
                           procedure set_text(p_text: String);
                           procedure set_search_text(p_search_text: String);
                           procedure set_word_character_string(p_word_character_string: String);

                         Public
                           constructor Create(Aowner: tComponent); Override;
                           procedure load_from_file(p_full_file_name: String);
                           Destructor Destroy; Override;
                         Published
                           Property Text: String write set_text;
                           Property SearchText: String write set_search_text;
                           Property WordCharacters: String read m_word_character_string write set_word_character_string;

                           Property Path: String read m_path write m_path;
                           Property FileName: String read m_file_name write m_file_name;
                       end; // tStearchMemo

  implementation
    Uses Windows, Messages, SysUtils
       ;

    type t_set_of_char= Set of Char;

    // function f_string_to_set_of_char(p_string: String): t_set_of_char;
    function f_string_to_set_of_char(p_string: String): t_set_of_char;
      var l_index, l_length: Integer;
          l_start_character: Char;

      procedure add_range;
        var l_character: Char;
        begin
          Inc(l_index);
          if (l_index<= l_length) and (p_string[l_index]= '.')
            then begin
                Inc(l_index);
                if (l_index<= l_length)
                  then begin
                      Result:= Result+ [l_start_character..p_string[l_index]];
                      Inc(l_index);
                    end
                  else ; // bug
              end
            else ; // bug
        end; // add_range

      procedure add_character;
        begin
          Result:= Result+ [p_string[l_index]];
          Inc(l_index);
        end; // add_character

      begin // f_string_to_set_of_char
        l_length:= Length(p_string);
        Result:= [];

        l_index:= 1;
        while l_index<= l_length do
        begin
          l_start_character:= p_string[l_index];
          Inc(l_index);
          if l_index<= l_length
            then
              if p_string[l_index]= '.'
                then add_range
                else add_character;
        end; // while l_index
      end; // f_string_to_set_of_char

    function f_display_character_set(p_set_of_char: t_set_of_char): String;
      begin
      end; // f_string_to_set_or_char

    // -- TSearchMemo

    constructor TSearchMemo.Create(Aowner: tComponent);

      procedure create_panel;
        const k_button_height= 25;
        var l_x: Integer;
        begin
          m_c_panel:= tPanel.Create(Self);

          with m_c_panel do
          begin
            Parent:= Self;
            Height:= k_button_height+ 2;
            Align:= alTop;
          end; // with m_c_panel

          l_x:= 15;

          m_c_previous_button:= tButton.Create(Self);
          with m_c_previous_button do
          begin
            Parent:= m_c_panel;
            Caption:= '<';
            SetBounds(l_x, 2, 20, k_button_height);

            Inc(l_x, m_c_previous_button.Width+ 5);
          end; // with m_c_previous_button

          m_c_search_edit:= tEdit.Create(Self);
          with m_c_search_edit do
          begin
            Parent:= m_c_panel;
            // Visible:= True;
            SetBounds(l_x, 2, 100, 30);

            OnKeyPress:= handle_richedit_keypress;
            OnKeyDown:= handle_memo_and_search_edit_keydown;

            Inc(l_x, Width+ 5);
          end; // with m_c_search_edit

          m_c_next_button:= tButton.Create(Self);
          with m_c_next_button do
          begin
            Parent:= m_c_panel;
            Caption:= '>';
            SetBounds(l_x, 2, 20, k_button_height);

            m_c_next_button.OnClick:= handle_next_click;

            Inc(l_x, m_c_next_button.Width+ 5);
          end; // with m_c_next_button

          m_c_line_number_label:= tLabel.Create(Self);
          with m_c_line_number_label do
          begin
            Parent:= m_c_panel;
            SetBounds(l_x, 2+ 2, 20, k_button_height);
            // Visible:= true;
            Caption:= 'line';

            Inc(l_x, m_c_line_number_label.Width+ 5);
          end; // with m_c_line_number_label

          m_c_debug_label:= tLabel.Create(Self);
          with m_c_debug_label do
          begin
            Parent:= m_c_panel;
            SetBounds(l_x, 2+ 2, 120, k_button_height);
            // Visible:= true;
            Caption:= 'debug';
            AutoSize:= True;

            Inc(l_x, m_c_line_number_label.Width+ 5);
          end; // with m_c_line_number_label
        end; // create_panel

      procedure create_richedit;
        begin
          m_c_richedit:= trichedit.Create(Self);
          with m_c_richedit do
          begin
            Parent:= Self;
            Align:= alClient;
            ScrollBars:= ssVertical;

            // "has no parent window"  HideSelection:= False;

(*
            OnKeyDown:= handle_richedit_and_search_edit_keydown;
            OnChange:= handle_richedit_change;
*)
          end; // with m_c_richedit
        end; // create_richedit

      begin // Create
        Inherited Create(Aowner);

        Align:= alClient;

        create_panel;
        create_richedit;

        m_word_character_string:= 'a..z A..Z';
        m_word_character_set:= f_string_to_set_of_char(m_word_character_string); // ['a'..'z', 'A'..'Z'];

        m_whole_word_color:= clRed;
        m_partial_word_color:= clBlue;
      end; // Create

    procedure TSearchMemo.set_text(p_text: String);
      begin
        m_c_richedit.Text:= p_text;
      end; // set_text

    procedure TSearchMemo.set_search_text(p_search_text: String);
      begin
        m_search_string:= p_search_text;
        m_c_search_edit.Text:= m_search_string;
      end; // set_search_text

    procedure TSearchMemo.set_word_character_string(p_word_character_string: String);
      begin
        m_word_character_string:= p_word_character_string;
        m_word_character_set:= f_string_to_set_of_char(m_word_character_string);
      end; // set_word_character_string

    // -- computations

    procedure TSearchMemo.load_from_file(p_full_file_name: String);
      begin
        m_c_richedit.Lines.LoadFromFile(p_full_file_name);
      end; // load_from_file

    procedure TSearchMemo.set_panel_color(p_color: Integer);
      begin
        m_c_panel.Color:= p_color;
      end; // set_panel_color

    function TSearchMemo.f_memo_line_of_index(p_text_index: Integer): Integer;
      var l_line_index: Integer;
          l_running_text_index: Integer;
      begin
        Result:= 0;
        l_running_text_index:= 0;
        with m_c_richedit do
          for l_line_index:= 0 to Lines.Count- 1 do
            if l_running_text_index> p_text_index
              then begin
                  Result:= l_line_index- 1;
                  Break;
                end
              else Inc(l_running_text_index, Length(Lines[l_line_index])+ 2);
      end; // f_memo_line_of_index

    procedure TSearchMemo.set_memo_top_line_index(p_top_line_index: Integer);
      var l_first_visible_line: Integer;
      begin
        with m_c_richedit do
        begin
          l_first_visible_line:= Perform(em_GetFirstVisibleLine, 0, 0);
          if l_first_visible_line<> p_top_line_index
            then Perform(em_LineScroll, 0, p_top_line_index- l_first_visible_line);
        end;
      end; // set_memo_top_line_index

    function TSearchMemo.f_is_whole_word: Boolean;
        // -- true if isolated word

      function f_start_at_beginning: Boolean;
          // -- the character before is not a word character
        begin
          Result:=
              (m_text_index<= 1)
            or
              not (m_text[m_text_index- 1] in m_word_character_set);

          if Result
            then m_c_debug_label.Caption:= 'beg '
            else m_c_debug_label.Caption:= 'not_beg ';

          m_c_debug_label.Caption:= m_c_debug_label.Caption+ '['+ m_text[m_text_index- 1]+ '] ';
        end; // f_start_at_beginning

      function f_end_at_end: Boolean;
        begin
          Result:=
              (m_text_index+ m_search_length- 1>= m_text_length)
            or
              not (m_text[m_text_index+ m_search_length- 1+ 1] in m_word_character_set);

          if Result
            then m_c_debug_label.Caption:= m_c_debug_label.Caption+ 'end'
            else m_c_debug_label.Caption:= m_c_debug_label.Caption+ 'not_end ';
          m_c_debug_label.Caption:= m_c_debug_label.Caption+ '['+ m_text[m_text_index+ m_search_length- 1+ 1]+ '] ';
        end; // f_end_at_end

      begin // f_is_whole_word
        Result:= f_start_at_beginning and f_end_at_end;
      end; // f_is_whole_word

    procedure TSearchMemo.search_next_occurence(p_color: Integer);
        // -- search string

      function f_search_rest(p_index: Integer): Boolean;
        var l_word_index: Integer;
        begin
          l_word_index:= 2;
          while (l_word_index<= m_search_length)
              and (p_index+ l_word_index<= m_text_length)
              and (m_search_string[l_word_index]= m_text[p_index+ l_word_index- 1]) do
            inc(l_word_index);

          Result:= l_word_index> Length(m_search_string);
        end; // f_search_rest

      var l_first_character: Char;

      begin // search_next_occurence
        with m_c_richedit do
        begin
          Inc(m_text_index, m_search_length);

          l_first_character:= m_search_string[1];

          while m_text_index< m_text_length do
          begin
            if m_text[m_text_index]= l_first_character
              then
                if f_search_rest(m_text_index)
                  then begin
                      // Inc(m_text_index, Length(m_search_string)- 1);
                      if p_color= clRed
                        then begin
                            SelStart:= m_text_index- 1;
                            SelLength:= m_search_length;
                            SelAttributes.Color:= clRed;
                            if f_is_whole_word
                              then SelAttributes.Color:= m_whole_word_color
                              else SelAttributes.Color:= m_partial_word_color;
                            SelAttributes.Style:= [fsBold];
                          end
                        else begin
                            set_memo_top_line_index(f_memo_line_of_index(m_text_index));
                            m_c_line_number_label.Caption:= IntToStr(f_memo_line_of_index(m_text_index));

                            SelStart:= m_text_index- 1;
(*
                            if m_do_hilite_result
                              then
*)
                                SelLength:= m_search_length;

                            if f_is_whole_word
                              then SelAttributes.Color:= m_whole_word_color
                              else SelAttributes.Color:= m_partial_word_color;

                            set_panel_color(clLime);
                          end;

                      Exit;
                    end;

            inc(m_text_index);
          end; // while pv_index

        end; // with m_c_richedit do
        set_panel_color(clRed);
      end; // search_next_occurence

    // -- events

    const k_return= Chr(13);

    procedure TSearchMemo.handle_richedit_keypress(p_c_sender: TObject; var pv_key: Char);
        // -- when hits Enter after changing the search text, start the search

      procedure search_all_occurences;
        begin
          // -- search_next_occurence always begins by incrementing by search length
          m_text_index:= 1- m_search_length;
          while m_text_index< m_text_length do
          begin
            search_next_occurence(clRed);
          end;
        end; // search_all_occurences

      procedure position_on_first_occurence;
        begin
          m_text_index:= 1- m_search_length;
          search_next_occurence(clBlack);
        end; // position_on_first_occurence

      begin // handle_richedit_keypress
        if pv_key= k_return
          then begin
              pv_key:= chr(0);
              m_search_string:= m_c_search_edit.Text;
              m_search_length:= Length(m_search_string);

              m_text:= m_c_richedit.Text;
              m_text_length:= Length(m_text);

              search_all_occurences;
              m_c_richedit.HideSelection:= False;
              position_on_first_occurence;
            end;
      end; // handle_richedit_keypress

    procedure TSearchMemo.handle_memo_and_search_edit_keydown(p_c_sender: TObject;
        var pv_scan_code: Word; p_shift_state: TShiftState);
      begin
        case pv_scan_code of
          vk_control : _m_did_hit_control:= True;
          vk_escape : begin
                        set_panel_color(clBtnFace);
                        m_c_richedit.SelLength:= 0;
                      end;
          Ord('L') : if _m_did_hit_control
                       then begin
                           // -- no beep. Does not work ?
                           pv_scan_code:= 0;

                           search_next_occurence(clBlack);
                           // -- do not cancel the control, if keeps finger
                           // --   on Ctrl for repeated "search next"
                           // m_did_hit_control:= False;
                         end;
          Ord('F') : if _m_did_hit_control
                       then begin
                           // -- no beep
                           pv_scan_code:= 0;

                           m_c_search_edit.SetFocus;
                           // -- do not cancel the control, if keeps finger
                           // --   on Ctrl for repeated "search next"
                           // m_did_hit_control:= False;
                         end;
          else
            _m_did_hit_control:= False;
        end; // case pv_scan_code
      end; // handle_richedit_and_search_edit_keydown

    procedure TSearchMemo.handle_next_click(p_c_sender: tObject);
      begin
        search_next_occurence(clBlack);
      end; // handle_next_click

    Destructor TSearchMemo.Destroy;
      begin
        Inherited;
      end; // Destroy

  end.
